@@ -12,17 +12,14 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import numpy as onp
-from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union
+from typing import Any, Callable, Dict, Sequence, Union
 
-import jax.numpy as np
+import jax.numpy as jnp
 
 from jax import core
 from jax.core import Trace, Tracer, new_master
 from jax import linear_util as lu
-from jax.util import partial, safe_map, wrap_name
-from jax.interpreters import xla
-from jax.interpreters import partial_eval as pe
+from jax.util import partial, safe_map
 
 import inspect
 from jax.api_util import (wraps, flatten_fun_nokwargs)
@@ -76,14 +73,14 @@ def _contains_query(vals, query):
   if isinstance(query, tuple):
     return map(partial(_contains_query, vals), query)
 
-  if np.isnan(query):
-    if np.any(np.isnan(vals)):
+  if jnp.isnan(query):
+    if jnp.any(jnp.isnan(vals)):
       raise FoundValue('NaN')
-  elif np.isinf(query):
-    if np.any(np.isinf(vals)):
+  elif jnp.isinf(query):
+    if jnp.any(jnp.isinf(vals)):
       raise FoundValue('Found Inf')
-  elif np.isscalar(query):
-    if np.any(vals == query):
+  elif jnp.isscalar(query):
+    if jnp.any(vals == query):
       raise FoundValue(str(query))
   else:
     raise ValueError('Malformed Query: {}'.format(query))
