@@ -13,11 +13,9 @@
 # limitations under the License.
 """Experimental module transforms JAX functions to be executed by TensorFlow."""
 
-import contextlib
 import functools
 import string
-import threading
-from typing import Any, Callable, Dict, Iterable, Sequence, Union
+from typing import Any, Callable, Dict, Iterable, Sequence, Tuple, Union
 
 import jax
 from jax import abstract_arrays
@@ -47,6 +45,8 @@ import tensorflow as tf  # type: ignore[import]
 # pylint: disable=g-direct-tensorflow-import
 from tensorflow.compiler.tf2xla.python import xla as tfxla  # type: ignore[import]
 from tensorflow.compiler.xla import xla_data_pb2  # type: ignore[import]
+from tensorflow.python.ops import control_flow_util  # type: ignore[import]
+from tensorflow.python.framework import ops  # type: ignore[import]
 
 # A value suitable in a TF tracing context: tf.Tensor, tf.Variable,
 # or Python scalar or numpy.ndarray. (A tf.EagerTensor is a tf.Tensor.)
@@ -80,13 +80,20 @@ def _tfval_remove_unit(args: Sequence[TfValOrUnit]) -> Sequence[TfVal]:
 def _tfval_add_unit(vals: Sequence[TfValOrUnit],
                     avals: Sequence[core.AbstractValue]) -> Sequence[TfValOrUnit]:
   """Turn regular TfVals into TfValOrUnit, based on expected abstract values.
-  This function is sometimes called with a mix of core.unit and tf.nan in places
-  of units.
+  When the aval is a unit, the corresponding value is either core.unit,
+  or an EagerTensor with the value NaN (we use tf.nan as a concrete TF value
+  for units, see _tfval_remove_unit) or may even be a Tensor if we are building
+  graphs and the NaN value is abstracted.
   """
   def add_unit(v: TfValOrUnit, aval: core.AbstractValue):
     if not core.skip_checks:
-      assert ((v is core.unit or tf.math.is_nan(v))
-              if aval is core.abstract_unit else _is_tfval(v))
+      if aval is core.abstract_unit:
+        if v is not core.unit:
+          assert isinstance(v, tf.Tensor)
+          if v.device:  # Only for EagerTensor
+            assert tf.math.is_nan(v)
+      else:
+        assert _is_tfval(v)
     return core.unit if aval is core.abstract_unit else v
   return util.safe_map(add_unit, vals, avals)
 
@@ -98,64 +105,18 @@ def _tfval_add_unit(vals: Sequence[TfValOrUnit],
 tf_impl: Dict[core.Primitive,
               Callable[..., Any]] = {}
 
-# TODO(necula): used for tests, until we handle all control-flow primitives
-class _JitState(threading.local):
-
-  def __init__(self):
-    super().__init__()
-    self.disable_jit = True
-
-_jit_state = _JitState()
-
-
-@contextlib.contextmanager
-def enable_jit():
-  """Temporarily allow JAX jit."""
-  old_state = _jit_state.disable_jit
-  _jit_state.disable_jit = False
-  try:
-    yield
-  finally:
-    _jit_state.disable_jit = old_state
-
-
-def disable_gradient(fun):
-  """Prevents the wrapped function from being differentiated."""
-
-  def grad_disabled(*dy, variables=None):
-    raise ValueError("jax2tf currently does not support gradients. Please "
-                     "reach out to tomhennigan@ if this feature is a blocker "
-                     "for you, we are working on it!")
-
-  is_tensor = lambda t: isinstance(t, (tf.Tensor, tf.Variable))
-
-  def wrapper(*args, **kwargs):
-    flat_args, treedef = jax.tree_flatten((args, kwargs))
-    tensor_idxs = [i for i, t in enumerate(flat_args) if is_tensor(t)]
-    tensor_args = [t for t in flat_args if is_tensor(t)]
-
-    @tf.custom_gradient
-    def inner_wrapper(*tensor_args):
-      flat_copy = list(flat_args)
-      for i, t in zip(tensor_idxs, tensor_args):
-        flat_copy[i] = t
-
-      args, kwargs = jax.tree_unflatten(treedef, flat_copy)
-      out = fun(*args, **kwargs)
-      return out, grad_disabled
-
-    return inner_wrapper(*tensor_args)
-
-  return wrapper
-
-
-def convert(fun):
+def convert(fun, with_gradient=False):
   """Transforms `fun` to be executed by TensorFlow.
 
   Args:
     fun: Function to be transformed. Its arguments and return value should be
       JAX arrays, or (nested) standard Python containers (tuple/list/dict)
       thereof.
+    with_gradient: if set, will add a tf.custom_gradient to the converted
+      function, by converting the ``jax.vjp(fun)``. Only first-order
+      differentiation is supported for now. If the converted function is
+      saved in a SavedModel, the custom gradients are currently lost and
+      an error will be raised if a gradient computation is attempted.
 
   Returns:
     A version of `fun` that expects TfVals as arguments (or
@@ -163,30 +124,59 @@ def convert(fun):
   """
   api._check_callable(fun)
 
-  @disable_gradient
-  def wrapped_fun(*args: TfValOrUnit) -> TfValOrUnit:
-    # TODO(necula): remove the jit disabling once we handle all control-flow.
-    # Disabling the jit helps to avoid some unsupported jax primitives.
-    # E.g. scan will be statically unrolled.
-    def doit():
-      f = lu.wrap_init(fun)
-      args_flat, in_tree = tree_util.tree_flatten((args, {}))
-      for a in args_flat:
-        if not _is_tfvalorunit(a):
-          msg = (f"Argument {a} of type {type(a)} of jax2tf.convert(f) should "
-                 "be NumPy array, scalar, tf.Variable, or tf.Tensor")
-          raise TypeError(msg)
-      flat_fun, out_tree = flatten_fun(f, in_tree)
-      out_flat = _interpret_fun(flat_fun, args_flat)
-      return tree_util.tree_unflatten(out_tree(), out_flat)
+  def converted_fun(*args: TfVal) -> TfVal:
+    # This function may take pytrees of TfVals. We can only set
+    # tf.custom_gradient on functions that take a flat argument list.
+    args_flat, in_tree = tree_util.tree_flatten((args, {}))
+    for a in args_flat:
+      if not _is_tfvalorunit(a):
+        msg = (f"Argument {a} of type {type(a)} of jax2tf.convert(f) should "
+               "be NumPy array, scalar, tf.Variable, or tf.Tensor")
+        raise TypeError(msg)
 
-    if _jit_state.disable_jit:
-      with jax.disable_jit():
-        return doit()
+    f = lu.wrap_init(fun)
+    # out_tree_thunk() will be the output tree, after running _interpret_fun.
+    flat_fun, out_tree_thunk = flatten_fun(f, in_tree)
+
+    # Prepare the grad_fn for tf.custom_gradient.
+    def converted_grad_fn(*out_cts_flat: TfVal, **kwargs):
+      # TODO(cl/318778369): change **kwargs with variables=None
+      variables = kwargs.get("variables", [])
+      if variables:
+        raise ValueError("Unexpected variables used in forward pass. "
+                         "This should not happen for first-order differentiation. "
+                         f"variables={variables}")
+
+      def fun_vjp_jax(args_jax, out_cts_jax):
+        # One may think that we can get the pullback while we are converting
+        # the main function in the first place. That is problematic, because the
+        # pullback may contain captured tracers from the conversion of the
+        # main function. Those tracers will confuse the conversion of the
+        # pullback. So, we construct the vjp anew.
+        _, pullback_jax = jax.vjp(fun, *args_jax)
+        return pullback_jax(out_cts_jax)
+      out_cts = tree_util.tree_unflatten(out_tree_thunk(), out_cts_flat)
+      in_cts = convert(fun_vjp_jax, with_gradient=False)(args, out_cts)
+      return in_cts
+
+    if with_gradient:
+      @tf.custom_gradient
+      def converted_fun_flat_with_custom_gradient(*args_flat: TfVal) -> TfVal:
+        return _interpret_fun(flat_fun, args_flat), converted_grad_fn
+
+      out_flat = converted_fun_flat_with_custom_gradient(*args_flat)
     else:
-      return doit()
+      out_flat_raw = _interpret_fun(flat_fun, args_flat)
+      message = ("The jax2tf-converted function does not support gradients. "
+                 "Use `with_gradient` parameter to enable gradients")
+      # We use PreventGradient, which is propagated through a SavedModel.
+      out_flat = [tf.raw_ops.PreventGradient(input=o, message=message)
+                  for o in out_flat_raw]
 
-  return wrapped_fun
+    out = tree_util.tree_unflatten(out_tree_thunk(), out_flat)
+    return out
+
+  return converted_fun
 
 # Internals
 
@@ -227,7 +217,7 @@ def _interpret_jaxpr(jaxpr: core.TypedJaxpr, *args: TfValOrUnit) -> Sequence[TfV
 
 
 def abstractify(t: Union[tf.Tensor, tf.Variable]):
-  return abstract_arrays.ShapedArray(tuple(t.shape), t.dtype.as_numpy_dtype)
+  return abstract_arrays.ShapedArray(tuple(t.shape), to_jax_dtype(t.dtype))
 
 # TODO(b/26854495): pylint doesn't understand slots and inheritance.
 # pylint: disable=assigning-non-slot
@@ -244,7 +234,7 @@ class TensorFlowTracer(core.Tracer):
       self.val = val
     elif isinstance(val, (tf.Tensor, tf.Variable)):
       aval: core.ShapedArray = abstractify(val)
-      if np.dtype(aval.dtype) != val.dtype.as_numpy_dtype:  # type: ignore
+      if np.dtype(aval.dtype) != to_jax_dtype(val.dtype):  # type: ignore
         # This is expected when JAX 64-bit is not enabled
         self.val = tf.cast(val, dtype=aval.dtype)
       else:
@@ -331,11 +321,19 @@ class TensorFlowTrace(core.Trace):
       msg = "TensorFlow interpretation rule for '{}' not implemented"
       raise NotImplementedError(msg.format(p))
 
+def to_tf_dtype(jax_dtype):
+  if jax_dtype == jnp.bfloat16:
+    return tf.bfloat16
+  else:
+    return tf.dtypes.as_dtype(jax_dtype)
+
+def to_jax_dtype(tf_dtype):
+  return jnp.bfloat16 if tf_dtype == tf.bfloat16 else tf_dtype.as_numpy_dtype
 
 def promote_types(*values):
   """Returns values casted to a common type using jnp.promote_types."""
-  dtype = tf.dtypes.as_dtype(functools.reduce(
-      jnp.promote_types, (v.dtype.as_numpy_dtype for v in values)))
+  dtype = to_tf_dtype(functools.reduce(
+      jnp.promote_types, (to_jax_dtype(v.dtype) for v in values)))
   return tuple(tf.cast(v, dtype) for v in values)
 
 
@@ -355,25 +353,23 @@ for unexpected in [
 
 # Primitives that are not yet implemented must be explicitly declared here.
 tf_not_yet_impl = [
-  ad.custom_lin_p,
-
-  lax.after_all_p, lax.all_to_all_p, lax.create_token_p, lax_fft.fft_p,
-  lax.igamma_grad_a_p, lax.infeed_p, lax.linear_solve_p, lax.outfeed_p,
-  lax.sort_p, lax.pmax_p, lax.pmin_p, lax.ppermute_p, lax.psum_p,
   lax.population_count_p, lax.reduce_p, lax.reduce_window_p, lax.rng_uniform_p,
   lax.select_and_gather_add_p, lax.select_and_scatter_p,
-  lax.top_k_p,
 
-  core.call_p,
+  lax.linear_solve_p,
   lax_linalg.cholesky_p, lax_linalg.eig_p, lax_linalg.eigh_p,
-  lax_linalg.lu_p, lax_linalg.qr_p, lax_linalg.svd_p,
+  lax_linalg.lu_p, lax_linalg.svd_p,
   lax_linalg.triangular_solve_p,
 
-  custom_derivatives.custom_jvp_call_jaxpr_p,
-  custom_derivatives.custom_vjp_call_jaxpr_p,
-
+  lax_fft.fft_p, lax.igamma_grad_a_p,
   random.random_gamma_p,
   lax.random_gamma_grad_p,
+
+  # Not high priority?
+  lax.after_all_p, lax.all_to_all_p, lax.create_token_p, lax.cummax_p, lax.cummin_p,
+  lax.infeed_p, lax.outfeed_p, lax.pmax_p, lax.pmin_p, lax.ppermute_p, lax.psum_p,
+
+  core.call_p,
   pe.remat_call_p,
   pxla.xla_pmap_p, pxla.axis_index_p,
 ]
@@ -434,6 +430,20 @@ tf_impl[lax.add_p] = wrap_binary_op(tf.math.add)
 tf_impl[lax.sub_p] = wrap_binary_op(tf.math.subtract)
 tf_impl[lax.mul_p] = wrap_binary_op(tf.math.multiply)
 
+def _xla_compile(func: Callable, *args: TfVal) -> TfVal:
+  """Ensure that the function is compiled with XLA.
+
+  This is needed to work around some bugs, e.g., without XLA
+  a certain TF op has different behavior than expected by JAX.
+  """
+  # Do not invoke XLA if we are already in an XLA context
+  in_xla_context = control_flow_util.GraphOrParentsInXlaContext(
+    ops.get_default_graph())
+  if in_xla_context:
+    return func(*args)
+  else:
+    res, = tf.xla.experimental.compile(func, args)
+    return res
 
 def _div(lhs, rhs):
   if lhs.dtype.is_integer:
@@ -542,7 +552,7 @@ tf_impl[lax.lt_p] = wrap_binary_op(tf.math.less)
 
 def _convert_element_type(operand, new_dtype, old_dtype):
   del old_dtype
-  return tf.dtypes.cast(operand, new_dtype)
+  return tf.dtypes.cast(operand, to_tf_dtype(new_dtype))
 tf_impl[lax.convert_element_type_p] = _convert_element_type
 
 
@@ -722,7 +732,9 @@ tf_impl[lax.slice_p] = _slice
 
 
 def _dynamic_slice(operand, *start_indices, slice_sizes=None):
-  return tf.slice(operand, tf.stack(start_indices), slice_sizes)
+  # See out-of-bounds index clamping comment for tf.gather
+  return _xla_compile(lambda o, s: tf.slice(o, s, slice_sizes),
+                      operand, tf.stack(start_indices))
 tf_impl[lax.dynamic_slice_p] = _dynamic_slice
 
 
@@ -749,6 +761,14 @@ tf_impl[lax.reduce_min_p] = (
 tf_impl[lax.reduce_or_p] = axes_to_axis(tf.reduce_any)
 tf_impl[lax.reduce_and_p] = axes_to_axis(tf.reduce_all)
 
+def _argminmax(fn, operand, axes, index_dtype):
+  axis, = axes
+  # TODO(phawkins): handle axes larger than 2^31.
+  return fn(operand, axis=axis, output_type=to_tf_dtype(index_dtype))
+
+tf_impl[lax.argmin_p] = functools.partial(_argminmax, tf.math.argmin)
+tf_impl[lax.argmax_p] = functools.partial(_argminmax, tf.math.argmax)
+
 
 _add_fn = tf.function(tf.math.add)
 _ge_fn = tf.function(tf.math.greater_equal)
@@ -761,11 +781,14 @@ tf_impl[lax.cumprod_p] = tf.math.cumprod
 
 
 def _reduce_window_shape(jax_f, operand, window_dimensions,
-                         window_strides, padding, input_shape=None):
+                         window_strides, padding, base_dilation,
+                         window_dilation, input_shape=None):
   """Shape inference function for reduce_window_{sum,min,max}."""
   params = dict(window_dimensions=window_dimensions,
                 window_strides=window_strides,
                 padding=padding,
+                base_dilation=base_dilation,
+                window_dilation=window_dilation,
                 input_shape=input_shape)
   try:
     out, = _infer_shape_jax(jax_f, operand, **params)
@@ -782,20 +805,20 @@ def _get_shape_from_tensor_or_array(x):
 
 
 def _reduce_window(jax_f, reducer, init_val, operand, window_dimensions,
-                   window_strides, padding, input_shape=None):
+                   window_strides, padding, base_dilation, window_dilation,
+                   input_shape=None):
   """TensorFlow implementation of reduce_window_{sum,min,max}."""
   del input_shape
   # TODO(tomhennigan): tf2xla should have a shape inference function.
   out_shape = _reduce_window_shape(jax_f, operand, window_dimensions,
-                                   window_strides, padding)
-  padding = lax.padtype_to_pads(_get_shape_from_tensor_or_array(operand),
-                                window_dimensions,
-                                window_strides, padding)
+                                   window_strides, padding, base_dilation,
+                                   window_dilation)
   a = tf.constant(0, operand.dtype)
   reducer_fn = reducer.get_concrete_function(a, a)
   out = tfxla.reduce_window(operand, tf.constant(init_val, operand.dtype),
                             reducer_fn, window_dimensions,
-                            window_strides, padding=padding)
+                            window_strides, base_dilations=base_dilation,
+                            window_dilations=window_dilation, padding=padding)
   out.set_shape(out_shape)
   return out
 # pylint: disable=protected-access
@@ -936,9 +959,12 @@ def _try_tf_gather(operand, start_indices, dimension_numbers, slice_sizes):
     return None
   # TODO: should we allow ourselves to add a reshape, or should we strictly
   #  convert 1:1, or go to TFXLA when not possible?
-  start_indices = tf.reshape(start_indices, start_indices.shape[0:-1])
-  return tf.gather(operand, start_indices, axis=axis, batch_dims=0)
-
+  start_indices_reshaped = tf.reshape(start_indices, start_indices.shape[0:-1])
+  # We do not use tf.gather directly because in the non-compiled version it
+  # rejects indices that are out of bounds, while JAX and XLA clamps them.
+  # We fix this by ensuring that we always compile tf.gather, to use XLA.
+  return _xla_compile(lambda o, s: tf.gather(o, s, axis=axis, batch_dims=0),
+                      operand, start_indices_reshaped)
 
 @functools.partial(bool_to_int8, argnums=0)
 def _gather(operand, start_indices, dimension_numbers, slice_sizes):
@@ -949,9 +975,11 @@ def _gather(operand, start_indices, dimension_numbers, slice_sizes):
   out_shape = _gather_shape(
       operand, start_indices, dimension_numbers, slice_sizes)
   proto = _gather_dimensions_proto(start_indices.shape, dimension_numbers)
-  out, = tf.xla.experimental.compile(
-      lambda o, s: tfxla.gather(o, s, proto, slice_sizes, False),
-      [operand, start_indices])
+  # We compile because without it we run into a TF bug:
+  # tfxla.gather fails on constant inputs with "must be a compile-time constant"
+  # b/153556869
+  out = _xla_compile(lambda o, s: tfxla.gather(o, s, proto, slice_sizes, False),
+                     operand, start_indices)
   out.set_shape(out_shape)
   return out
 tf_impl[lax.gather_p] = _gather
@@ -977,18 +1005,21 @@ def _scatter_shape(operand, scatter_indices, updates, dimension_numbers):
 
 @functools.partial(bool_to_int8, argnums=(1, 3))
 def _scatter(update_computation, operand, scatter_indices, updates,
-             update_jaxpr, update_consts, dimension_numbers):
+             update_jaxpr, update_consts, dimension_numbers,
+             indices_are_sorted, unique_indices):
   """Tensorflow implementation of scatter with an update computation."""
-  del update_jaxpr, update_consts
+  del update_jaxpr, update_consts, unique_indices
   out_shape = _scatter_shape(operand, scatter_indices, updates,
                              dimension_numbers)
   proto = _scatter_dimensions_proto(scatter_indices.shape, dimension_numbers)
   o_spec = tf.TensorSpec(None, dtype=operand.dtype)
   xla_update_computation = (
       tf.function(update_computation).get_concrete_function(o_spec, o_spec))
-  out, = tf.xla.experimental.compile(
-      lambda o, s, u: tfxla.scatter(o, s, u, xla_update_computation, proto),
-      [operand, scatter_indices, updates])
+  # We compile due to TF bug, see comment on gather
+  out = _xla_compile(
+    lambda o, s, u: tfxla.scatter(o, s, u, xla_update_computation, proto,
+                                  indices_are_sorted=indices_are_sorted),
+    operand, scatter_indices, updates)
   out.set_shape(out_shape)
   return out
 
@@ -1090,6 +1121,73 @@ def _scan(*tf_args : TfValOrUnit, **kwargs) -> Sequence[TfValOrUnit]:
 
 tf_impl[lax.scan_p] = _scan
 
+def _top_k(operand: TfVal, k: int) -> Tuple[TfVal, TfVal]:
+  # Some types originally incompatible with tf.math.top_k can be promoted
+  # to a compatible type without loss of precision.
+  def promote_tf_dtype(tf_dtype):
+    if tf_dtype in [tf.bool, tf.uint8, tf.uint16]:
+      return tf.uint32
+    if tf_dtype in [tf.int8, tf.int16]:
+      return tf.int32
+    if tf_dtype is tf.float16:
+      return tf.float32
+    return None
+
+  conversion_dtype = promote_tf_dtype(operand.dtype)
+  if conversion_dtype:
+    values, indices = tf.math.top_k(tf.dtypes.cast(operand, conversion_dtype), k=k, sorted=True)
+    return tf.dtypes.cast(values, operand.dtype), indices
+  else:
+    return tf.math.top_k(operand, k=k, sorted=True)
+
+tf_impl[lax.top_k_p] = _top_k
+
+def _sort(*operand: TfVal, dimension: int, is_stable: bool, num_keys: int) -> Tuple[TfVal, ...]:
+  if num_keys != 1:
+    raise NotImplementedError("TODO: multiple keys")
+  if len(operand) > 2:
+    raise NotImplementedError("TODO: handle > 2 tensors")
+  if is_stable:
+    raise NotImplementedError("TODO: implement stable version of XlaSort")
+  if dimension == len(operand[0].shape) - 1:
+    if len(operand) == 2:
+      return tuple(tfxla.key_value_sort(operand[0], operand[1]))
+    else:
+      return (tfxla.sort(operand[0]),)
+  else:
+    raise NotImplementedError("TODO: implement XlaSort for all axes")
+
+tf_impl[lax.sort_p] = _sort
+
+def _qr(operand, full_matrices):
+  return tf.linalg.qr(operand, full_matrices=full_matrices)
+
+tf_impl[lax_linalg.qr_p] = _qr
+
+def _custom_jvp_call_jaxpr(*args: TfValOrUnit,
+                           fun_jaxpr: core.TypedJaxpr,
+                           jvp_jaxpr_thunk: Callable) -> Sequence[TfValOrUnit]:
+  # TODO(necula): ensure that there is no AD transformation in scope
+  res = _interpret_jaxpr(fun_jaxpr, *args)
+  return _tfval_add_unit(res, fun_jaxpr.out_avals)
+
+tf_impl[custom_derivatives.custom_jvp_call_jaxpr_p] = _custom_jvp_call_jaxpr
+
+
+def _custom_vjp_call_jaxpr(*args: TfValOrUnit,
+                           fun_jaxpr: core.TypedJaxpr,
+                           **_) -> Sequence[TfValOrUnit]:
+  # TODO(necula): ensure that there is no AD transformation in scope
+  res = _interpret_jaxpr(fun_jaxpr, *args)
+  return _tfval_add_unit(res, fun_jaxpr.out_avals)
+
+tf_impl[custom_derivatives.custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr
+
+def _custom_lin(*args: TfValOrUnit, **_) -> Sequence[TfValOrUnit]:
+  raise TypeError("can't apply forward-mode autodiff (jvp) to a custom_vjp "
+                  "function.")
+
+tf_impl[ad.custom_lin_p] = _custom_lin
 
 def _register_checkpoint_pytrees():
   """Registers TF custom container types as pytrees."""
