@@ -18,10 +18,9 @@ from typing import Callable, Dict
 
 from .. import core
 from .. import linear_util as lu
-from ..core import Trace, Tracer, Primitive, new_master
+from ..core import Trace, Tracer, new_master
 from ..abstract_arrays import ShapedArray, raise_to_shaped
 from ..util import safe_map, safe_zip, unzip2, unzip3
-from . import partial_eval as pe
 
 map = safe_map
 zip = safe_zip
@@ -111,8 +110,6 @@ class PapplyTrace(Trace):
       return PapplyTracer(self, name, size, val_out, axis_out)
 
   def process_call(self, call_primitive, f: lu.WrappedFun, tracers, params):
-    if call_primitive in pe.map_primitives:
-      return self.process_map(call_primitive, f, tracers, params)
     names, vals, axes = unzip3((t.name, t.val, t.axis) for t in tracers)
     if all(axis is not_sharded for axis in axes):
       return call_primitive.bind(f, *vals, **params)
@@ -133,8 +130,5 @@ class PapplyTrace(Trace):
       return PapplyTracer(trace, name, size, x, axis)
     return val, todo
 
-  def process_map(self, map_primitive, f :lu.WrappedFun, tracers, params):
-    raise NotImplementedError  # TODO(mattjj,frostig)
-
 
 papply_primitive_rules: Dict[core.Primitive, Callable] = {}
